# Base Data
TIME_SERIES = 'TimeSeries'
WATER_LEVEL = 'WaterLevel'
PUMP = 'Pump'
BOX_CULVERT = 'BoxCulvert'
VALVE_OVERFLOW = 'ValveOverflow'
CUSTOM = 'Custom'
CAPACITY = 'Capacity'
INTERVAL = 'Interval'

# Q Out Data
OUTFLOW = 'Outflow'
OUT_BOX_CULVERT = f'Out.{BOX_CULVERT}'
OUT_VALVE_OVERFLOW = f'Out.{VALVE_OVERFLOW}'

# Q In Data
INFLOW = 'Inflow'